import os
import logging

USER_AGENT = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2883.95 Safari/537.36'

logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("chardet").setLevel(logging.WARNING)
logging.getLogger("requests").setLevel(logging.WARNING)
logging.basicConfig(level=logging.DEBUG, format='%(asctime)s - %(name)s - %(levelname)s: %(message)s')
LOGGER = logging.getLogger('search_engine')
SEARCH_ENGINE_SPIDER_DIR = os.path.realpath(os.path.dirname(__file__))
DATA_DIR = os.path.join(SEARCH_ENGINE_SPIDER_DIR, "data")


def assert_exists(file_name):
    if os.path.exists(file_name):
        return True
    print("not exists: " + file_name)

def file2list(file_name):
    assert_exists(file_name)
    with open(file_name) as f:
        lines = [line.strip() for line in f]
    return lines



# -*- coding:utf-8 -*-

import importlib
import itertools
import json
import os
import random
import sys

from numpy import median

from .search_engine import urlparse

from .config import DATA_DIR, SEARCH_ENGINE_SPIDER_DIR, file2list


class SearchEngineManager():
    '''
    search engine manager
    '''

    def __init__(self):
        self.user_agents_file = os.path.join(DATA_DIR, "user_agents.txt")
        self.user_agents = file2list(self.user_agents_file)
        self.search_engine_dir_name = "search_engine"
        self.search_engine_dir = os.path.join(SEARCH_ENGINE_SPIDER_DIR,
                                              self.search_engine_dir_name)
        self.search_engine_dict = dict()
        self.load_search_engine()

    def load_search_engine(self):
        '''
        from SEARCH_ENGINE_SPIDER_DIR/search_engine find search_engine
        '''
        for file_name in os.listdir(self.search_engine_dir):
            module_name, ext = os.path.splitext(file_name)
            if ext != ".py" or not module_name:
                continue
            module = importlib.import_module(".".join(
                (os.path.basename(SEARCH_ENGINE_SPIDER_DIR),
                 self.search_engine_dir_name, module_name)))
            search_engine_class = vars(module).get('LoadSearchEngine')
            if search_engine_class:
                search_engine = search_engine_class()
                self.search_engine_dict[
                    search_engine.get_name()] = search_engine

    def random_user_agent(self):
        '''
        random user agent
        '''
        return random.choice(self.user_agents)

    def get_search_engine(self, name):
        '''
        from name get search_engine object
        '''
        return self.search_engine_dict.get(name)

    def must_get_search_engine(self, name):
        '''
        if have not search_engine object, exit
        '''
        search_engine = self.get_search_engine(name)
        if search_engine:
            return search_engine
        print("没有该搜索引擎: " + name)
        exit(1)
        return False

    def print_search_engine(self):
        '''
        print exist search engine
        '''
        for search_engine_name in self.search_engine_dict:
            print(search_engine_name)

    def product_urls(self, engine_name, *url_file_list):
        '''
        from multi search_word file get search_url
        '''
        assert len(url_file_list) >= 1
        search_engine = self.must_get_search_engine(engine_name)
        for words in itertools.product(*map(file2list, url_file_list)):
            query_word = " ".join(words)
            print(search_engine.get_url(query_word))

    @staticmethod
    def extract_text(search_engine,
                     content,
                     max_num,
                     first_word=None,
                     end_word=None):
        '''
        from text extract search results
        '''
        results = list()
        for index, item in enumerate(search_engine.content2items(content)):
            if index >= max_num:
                continue
            results.append(
                list(
                    filter(None,
                           (first_word, item['url'], index + 1, end_word))))
        return results

    def extract_html(self, search_engine, html_path, max_num):
        '''
        from html_path get search_results
        '''
        with open(html_path) as html_f:
            return self.extract_text(search_engine, html_f.read(), max_num)

    def extract_json(self, search_engine, json_path, max_num):
        '''
        from json_path get search_results
        '''
        with open(json_path) as json_f:
            j = json.loads(json_f.read())
            url = j["url"]
            content = j["content"]
            query = search_engine.get_query(url)
            return self.extract_text(search_engine, content, max_num, query)

    def print_extract_html(self, engine_name, html_path, max_num=10000):
        '''
        from html_path get search_results and print
        '''
        max_num = int(max_num)
        search_engine = self.must_get_search_engine(engine_name)
        if os.path.isdir(html_path):
            for html_file in os.listdir(html_path):
                html_file = os.path.join(html_path, html_file)
                self.print_extract_html(engine_name, html_file, max_num)
        else:
            for item in self.extract_html(search_engine, html_path, max_num):
                print("\t".join(map(str, item)))

    def print_extract_json(self, engine_name, json_path, max_num=10000):
        '''
        from json_path get search_results and print
        '''
        max_num = int(max_num)
        search_engine = self.must_get_search_engine(engine_name)
        if os.path.isdir(json_path):
            for json_file in os.listdir(json_path):
                json_file = os.path.join(json_path, json_file)
                self.print_extract_json(engine_name, json_file, max_num)
        else:
            for item in self.extract_json(search_engine, json_path, max_num):
                print("\t".join(map(str, item)))

    @staticmethod
    def recovery_rate(local_site_file, json_result_file, max_num=10000):
        '''
        calculate site recovery rate
        '''
        local_site_set = set(file2list(local_site_file))
        search_site_dict = dict()
        with open(json_result_file) as json_result_f:
            for line in json_result_f:
                _, url, index = line.split("\t")
                if index > max_num:
                    continue
                parsed = urlparse(url)
                search_site_dict[parsed.hostname] = search_site_dict.get(
                    parsed.hostname, 0) + 1
        results = list()
        for site, items in search_site_dict.items():
            results.append([site, items, int(site in local_site_set)])
        results.sort(key=lambda x: x[1], reverse=True)
        for _ in results:
            print(" ".join(map(str, _)))

    @staticmethod
    def get_calc_func(calc_type):
        '''
        get calc func
        '''
        if calc_type == "len":
            calc_func = len
        elif calc_type == "median":
            calc_func = median
        elif calc_type == "reciprocal_median":
            calc_func = lambda x: median(map(lambda i: 1.0 / i, x))
        elif calc_type == "reciprocal_sum":
            calc_func = lambda x: sum(map(lambda i: 1.0 / i, x))
        else:
            print("error: calc_type")
            return False
        return calc_func

    @staticmethod
    def get_combination_word(*file_list, **kwargs):
        '''
        from multi file get combination words
        words_handle arg handle each words
        '''
        words_handle = kwargs.get("words_handle")
        if not words_handle:
            words_handle = lambda x: x
        combination_words = list()
        for words in itertools.product(*map(file2list, file_list)):
            if words:
                combination_words.append(words_handle(words))
        return combination_words

    def get_seed_value(self, calc_type, json_result_file, max_num,
                       *end_exclude_word_file_list):
        '''
        calculate seed value
        support len, median, reciprocal_median, reciprocal_sum
        '''
        calc_func = self.get_calc_func(calc_type)
        if not calc_func:
            return False

        exclude_words = self.get_combination_word(
            *end_exclude_word_file_list, words_handle=lambda x: (x, len(x)))
        exclude_words.sort(key=lambda x: x[1], reverse=True)
        search_word_dict = dict()
        '''
        {
            search_word: {
                seed: {
                    "value": 0,
                    "index_list": [index1, index2]
                }
            }
        }
        '''
        with open(json_result_file) as json_result_f:
            for _ in json_result_f:
                search_word, seed, index = _.strip().split()
                index = int(index)
                if index > max_num:
                    continue
                for exclude_word, exclude_word_len in exclude_words:
                    if search_word.endswith(exclude_word):
                        search_word = search_word[:-exclude_word_len].strip()
                        break
                search_word_obj = search_word_dict.get(search_word)
                if search_word_obj is None:
                    search_word_obj = dict()
                    search_word_dict[search_word] = search_word_obj
                seed_obj = search_word_obj.get(seed)
                if seed_obj is None:
                    search_word_obj[seed] = {
                        "value": 0,
                        "index_list": list(),
                    }
                seed_obj["index_list"].append(index)
            for search_word, search_word_obj in search_word_dict.items():
                for seed, seed_obj in search_word_obj.items():
                    index_list = seed_obj["index_list"]
                    index_list.sort()
                    seed_obj["value"] = calc_func(index_list)
        return search_word_dict

    def seed_value(self, calc_type, json_result_file, max_num, output_dir,
                   *end_exclude_word_file_list):
        '''
        calculate seed value and print
        support len, median, reciprocal_median, reciprocal_sum
        '''
        search_word_dict = self.get_seed_value(
            calc_type, json_result_file, max_num * end_exclude_word_file_list)
        if not search_word_dict:
            return False
        for search_word, search_word_obj in search_word_dict.items():
            seed_obj_list = list(search_word_obj.items())
            seed_obj_list.sort(key=lambda x: x[1]["value"], reverse=True)
            if output_dir == "-":
                output_file = sys.stdout
            else:
                output_file = open(os.path.join(output_dir, search_word), "w")
            for seed, seed_obj in seed_obj_list:
                seed_value = seed_obj["value"]
                index_list = seed_obj["index_list"]
                output_file.write(seed)
                output_file.write("\t")
                output_file.write(str(seed_value))
                output_file.write("\t")
                output_file.write(",".join(map(str, index_list)))
                output_file.write("\n")

            if output_file != sys.stdout:
                output_file.close()
        return True

    def get_site_value(self, calc_type, seed_calc_type, json_result_file,
                       max_num, *end_exclude_word_file_list):
        '''
        calculate site value
        support len, median, reciprocal_median, reciprocal_sum
        '''
        search_word_dict = self.get_seed_value(
            seed_calc_type, json_result_file,
            max_num * end_exclude_word_file_list)
        if not search_word_dict:
            return False
        calc_func = self.get_calc_func(calc_type)
        if not calc_func:
            return False
        search_word_site_dict = dict()
        for search_word, search_word_obj in search_word_dict.items():
            site_value_dict = dict()
            search_word_site_dict[search_word] = site_value_dict
            for seed, seed_obj in search_word_obj.item():
                hostname = urlparse(seed).hostname
                site_value_obj = site_value_dict.get(hostname)
                if not site_value_obj:
                    site_value_obj[hostname] = {
                        "value": 0,
                        "seed_value_list": list()
                    }
                site_value_obj[hostname]["seed_value_list"].append(
                    seed_obj["value"])
        for search_word, site_value_dict in search_word_site_dict.items():
            for site_value_obj in site_value_obj.values():
                site_value_obj["seed_value_list"].sort()
                site_value_obj["value"] = calc_func(
                    site_value_obj["seed_value_list"])

        return search_word_site_dict

    def site_value(self, calc_type, seed_calc_type, json_result_file, max_num,
                   output_dir, *end_exclude_word_file_list):
        '''
        calculate site value and print
        support len, median, reciprocal_median, reciprocal_sum
        '''
        search_word_site_dict = self.get_site_value(
            calc_type, seed_calc_type, json_result_file, max_num,
            *end_exclude_word_file_list)
        if not search_word_site_dict:
            return False
        for search_word, site_value_dict in search_word_site_dict.items():
            site_value_list = list(site_value_dict.items())
            site_value_list.sort(key=lambda x: x[1]["value"], reverse=True)
            with open(os.path.join(output_dir, search_word), "w") as output_f:
                for site, site_obj in site_value_list:
                    output_f.write(site)
                    output_f.write("\t")
                    output_f.write(site_obj["value"])
                    output_f.write("\n")
        return True



from .search_engine import SearchEngine, urlparse, parse_qs, quote_plus

from pyquery import PyQuery as pq
from bs4 import BeautifulSoup


class Bing(SearchEngine):
    def __init__(self):
        super(SearchEngine, self).__init__()
        self.URL = "https://{domain}/search?q={query}&first={start}"

    def content2items(self, content):
        soup = BeautifulSoup(content,"html.parser")
        results = soup.find(id="b_results")

        for li in results.children:
            if li.name != "li":
                continue
            h2 = li.find("h2")
            if not h2:
                continue
            url = li.a.get("href")
            if url and (url.startswith("http://") or url.startswith("https://")):
                title = h2.get_text()
                d = {
                    "url": url,
                    "title": title,
                    'text': '',
                }
                yield d

    def get_name(self):
        return 'bing'

    def get_domain(self):
        return "www.bing.com"

    def get_url(self, query, **kwargs):
        query = quote_plus(query)
        start = kwargs.get("start", 1)
        domain = kwargs.get("domain", self.get_domain())
        url=self.URL.format(domain=domain, query=query, start=start)
        return url

    def filter_link(self, link):
        """
        Returns None if the link doesn't yield a valid result.
        Token from https://github.com/MarioVilas/google
        :return: a valid result
        """
        try:
            # Valid results are absolute URLs not pointing to a Google domain
            # like images.google.com or googleusercontent.com
            o = urlparse(link, 'http')
            if o.netloc:
                return link
            # Decode hidden URLs.
            if link.startswith('/url?'):
                link = parse_qs(o.query)['q'][0]
                # Valid results are absolute URLs not pointing to a Google domain
                # like images.google.com or googleusercontent.com
                o = urlparse(link, 'http')
                if o.netloc:
                    return link
        # Otherwise, or on error, return None.
        except Exception as e:
            raise e

LoadSearchEngine=Bing



from .search_engine import SearchEngine, urlparse, parse_qs, quote_plus

from pyquery import PyQuery as pq


class Google(SearchEngine):
    def __init__(self):
        super(SearchEngine, self).__init__()
        self.URL = "https://{domain}/search?q={query}&btnG=Search&gbv=1&num={num}&start={start}"
    def content2items(self, content):
        pq_content = self.pq_html(content)
        for item in pq_content('div.g').items():
            result = {}
            result['title'] = item('h3.r>a').eq(0).text()
            href = item('h3.r>a').eq(0).attr('href')
            if href:
                url = self.filter_link(href)
                if not url:
                    continue
                result['url'] = url
            else:
                continue
            text = item('span.st').text()
            result['text'] = text
            yield result

    def get_name(self):
        return 'google'

    def get_domain(self):
        return "www.google.com"

    def get_url(self, query, **kwargs):
        query = quote_plus(query)
        start = kwargs.get("start", 0)
        domain = kwargs.get("domain", self.get_domain())
        num = kwargs.get("num", 50)
        url=self.URL.format(domain=domain, query=query, num=num, start=start)
        return url

    def filter_link(self, link):
        """
        Returns None if the link doesn't yield a valid result.
        Token from https://github.com/MarioVilas/google
        :return: a valid result
        """
        try:
            # Valid results are absolute URLs not pointing to a Google domain
            # like images.google.com or googleusercontent.com
            o = urlparse(link, 'http')
            if o.netloc:
                return link
            # Decode hidden URLs.
            if link.startswith('/url?'):
                link = parse_qs(o.query)['q'][0]
                # Valid results are absolute URLs not pointing to a Google domain
                # like images.google.com or googleusercontent.com
                o = urlparse(link, 'http')
                if o.netloc:
                    return link
        # Otherwise, or on error, return None.
        except Exception as e:
            raise e

LoadSearchEngine=Google



# -*- coding:utf-8 -*-

import os
import abc
import six
from pyquery import PyQuery as pq
from . import quote_plus, urlparse, parse_qs


SEARCH_ENGINE_DATA_DIR = os.path.join(os.path.dirname(__file__), "data")

@six.add_metaclass(abc.ABCMeta)
class SearchEngine():
    def __init__(self):
        pass

    def init(self):
        self.temp_data_dir = os.path.join(os.path.dirname(__file__), "temp_data")
        if not os.path.exists(self.temp_data_dir):
            os.mkdir(self.temp_data_dir)

        basename = os.path.basename(self.get_domain_file)

    def pq_html(self, content):
        """
        Parsing HTML by pyquery
        :param content: HTML content
        :return:
        """
        return pq(content)

    def get_domain_file(self, dir_path=SEARCH_ENGINE_DATA_DIR):
        """
        return: domain_file
        可以在SEARCH_ENGINE_DATA_DIR下面放入{engine_name}.txt的文件，每一行是一个域名
        """
        domain_file = os.path.join(dir_path, self.get_name()) + ".txt"
        if os.path.exists(domain_file):
            return domain_file
        return ''

    def get_query(self, url):
        """
        从url逆推出query
        """
        u = urlparse(url)
        return parse_qs(u.query)["q"][0]

    @abc.abstractmethod
    def get_name(self):
        """
        return: 搜索引擎名字(string)
        """
        pass

    @abc.abstractmethod
    def get_domain(self):
        """
        return: 搜索引擎默认域名(string)
        """
        pass

    @abc.abstractmethod
    def get_url(self, query, **kargs):
        """
        功能: 根据参数构造url
        必须:
        query: 要查询的词条(string)
        domain: 指定domain(默认提供)
        可选:
        start: 起始词条索引(默认0)，不同搜索引擎会根据自身情况适当修改
        num: 要查询的结果个数(int)，google有效
        return: url(string)
        """
        pass

    @abc.abstractmethod
    def content2items(self, content):
        """
        功能: 解析html源码生成搜索结果构建的数组
        content: html源码(string)
        return: 搜索结果(list[{
            "title": "",
            "url": "",
            "text": ""
        }])
        """
        pass



'''
search engine spider
'''

from setuptools import find_packages, setup

setup(
    name='search_engine_spider',
    version='0.1.3',
    description="A search engine spider",
    install_requires=[
        'pyquery>=1.2.17', 'requests>=2.12.4', 'cchardet', 'zcommand', 'bs4'
    ],
    author='izhengzhixian',
    author_email='izhengzhixian@gmail.com',
    platforms='any',
    packages=find_packages(),
    package_data={
        'search_engine_spider': ['data/*.txt'],
        'search_engine_spider.search_engine': ['data/*.txt']
    },
    scripts=['bin/search_engine_spider'],
)
